/**
 * Licensed under a dual GPL/BSD license.  (See LICENSE file for more info.)
 *
 * \file xsupconfig_parse_eap_tls.c
 *
 * \author chris@open1x.org
 *
 * $Id: xsupconfig_parse_eap_tls.c,v 1.1.2.18 2007/07/11 01:12:34 chessing Exp $
 * $Date: 2007/07/11 01:12:34 $
 **/

#include <stdio.h>

#ifndef WINDOWS
#include <stdint.h>
#include <strings.h>
#endif

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <string.h>

#include "xsupconfig_structs.h"
#include "../../src/xsup_common.h"
#include "xsupconfig.h"
#include "xsupconfig_parse.h"
#include "xsupconfig_parse_eap_tls.h"
#include "xsupconfig_common.h"

void *xsupconfig_parse_eap_tls(void **attr, xmlNodePtr node)
{
  struct config_eap_method *meth;

  meth = (*attr);

  if (meth == NULL)
  {
	meth = xsupconfig_alloc_method(meth, "EAP-TLS");
	(*attr) = meth;
  }
  else
  {
	  meth = xsupconfig_alloc_method(meth, "EAP-TLS");
  }

  if (meth == NULL) return NULL;

#ifdef PARSE_DEBUG
  printf("Parsing method 'EAP-TLS'.\n");
#endif

  meth->method_num = EAP_TYPE_TLS;
  meth->method_data = malloc(sizeof(struct config_eap_tls));
  if (meth->method_data == NULL)
    {
      printf("Couldn't allocate memory to store EAP-TLS data."
	     "  (Line %ld)\n", xsupconfig_parse_get_line_num());
      exit(2);
    }

  memset(meth->method_data, 0x00, sizeof(struct config_eap_tls));

  return meth->method_data;
}

void *xsupconfig_parse_eap_tls_user_cert(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS User Certificate : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->user_cert = NULL;
	}
	else
	{
		tls->user_cert = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_crl_dir(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS CRL Directory : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->crl_dir = NULL;
	}
	else
	{
		tls->crl_dir = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_user_key_file(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS User Key File : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->user_key = NULL;
	}
	else
	{
		tls->user_key = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_user_key_pass(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS User Key Password : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->user_key_pass = NULL;
	}
	else
	{
		tls->user_key_pass = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_enc_user_key_pass(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls = NULL;
  char *value = NULL;
  int size = 0;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS (Encrypted) User Key Password : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		return tls;
	}

  if (pwcrypt_decrypt(value, strlen(value), &tls->user_key_pass, &size) != 0)
  {
	  free(value);
	  tls->user_key_pass = NULL;
	  return tls;
  }

  if ((tls->user_key_pass != NULL) && (strlen(tls->user_key_pass) == 0))
  {
	  FREE(tls->user_key_pass);
  }

  free(value);

  return tls;
}

void *xsupconfig_parse_eap_tls_session_resume(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  uint8_t result;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS Session Resumption : %s\n", value);
#endif

  result = xsupconfig_common_yesno(value);

  if (result == 1)
    {
      tls->session_resume = TRUE;
    }
  else if (result == 0)
    {
      tls->session_resume = FALSE;
    }
  else
    {
      xsupconfig_common_log("Invalid value was passed for 'Session_Resume'!  Will use the "
	     "default value of no.  (Line %ld)\n", xsupconfig_parse_get_line_num());
      tls->session_resume = FALSE;
    }

  FREE(value);

  return tls;
}

void *xsupconfig_parse_eap_tls_chunk_size(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);
  
  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("Chunk Size : %s\n", value);
#endif

  if (xsupconfig_common_is_number(value) == 0)
    {
      xsupconfig_common_log("Value assigned to 'Chunk_Size' is not a number!  Using "
	     "default!  (Line %ld)\n", xsupconfig_parse_get_line_num());
    }
  else
    {
      tls->chunk_size = atoi(value);
    }

  FREE(value);

  return tls;
}

void *xsupconfig_parse_eap_tls_random_file(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

  tls = (*attr);

#ifdef PARSE_DEBUG
  printf("TLS Random File : %s\n", value);
#endif

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->random_file = NULL;
	}
	else
	{
		tls->random_file = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_engine_id(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

#ifdef PARSE_DEBUG
  printf("TLS Engine ID : %s\n", value);
#endif

  tls = (*attr);

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->sc.engine_id = NULL;
	}
	else
	{
		tls->sc.engine_id = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_trusted_server(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

#ifdef PARSE_DEBUG
  printf("Trusted Server : %s\n", value);
#endif

  tls = (*attr);

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->trusted_server = NULL;
	}
	else
	{
		tls->trusted_server = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_opensc_path(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

#ifdef PARSE_DEBUG
  printf("OpenSC Path : %s\n", value);
#endif

  tls = (*attr);

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->sc.opensc_so_path = NULL;
	}
	else
	{
		tls->sc.opensc_so_path = value;
	}

  return tls;
}

void *xsupconfig_parse_eap_tls_key_id(void **attr, xmlNodePtr node)
{
  struct config_eap_tls *tls;
  char *value;

  value = xmlNodeGetContent(node);

#ifdef PARSE_DEBUG
  printf("Key ID : %s\n", value);
#endif

  tls = (*attr);

	if ((value == NULL) || (strlen(value) == 0))
	{
		free(value);
		tls->sc.key_id = NULL;
	}
	else
	{
		tls->sc.key_id = value;
	}

  return tls;
}

parser eap_tls[] = {
  {"User_Certificate", NULL, FALSE, &xsupconfig_parse_eap_tls_user_cert},
  {"CRL_Directory", NULL, FALSE, &xsupconfig_parse_eap_tls_crl_dir},
  {"User_Key_File", NULL, FALSE, &xsupconfig_parse_eap_tls_user_key_file},
  {"User_Key_Password", NULL, FALSE, &xsupconfig_parse_eap_tls_user_key_pass},
  {"Encrypted_User_Key_Password", NULL, FALSE, &xsupconfig_parse_eap_tls_enc_user_key_pass},
  {"Session_Resume", NULL, FALSE, &xsupconfig_parse_eap_tls_session_resume},
  {"Chunk_Size", NULL, FALSE, &xsupconfig_parse_eap_tls_chunk_size},
  {"Random_File", NULL, FALSE, &xsupconfig_parse_eap_tls_random_file},
  {"Engine_ID", NULL, FALSE, &xsupconfig_parse_eap_tls_engine_id},
  {"OpenSC_Lib_Path", NULL, FALSE, &xsupconfig_parse_eap_tls_opensc_path},
  {"Key_ID", NULL, FALSE, &xsupconfig_parse_eap_tls_key_id},
  {"Trusted_Server", NULL, FALSE, &xsupconfig_parse_eap_tls_trusted_server},
  {"Type", NULL, FALSE, xsupcommon_do_nothing},

  // Add Certificate_ID and Root_Certificate_ID if they are ever implemented.

  {NULL, NULL, FALSE, NULL}};
